@@ -137,6 +137,8 @@ void QLowEnergyControllerPrivateAndroid::init()
                 this, &QLowEnergyControllerPrivateAndroid::descriptorWritten);
         connect(hub, &LowEnergyNotificationHub::characteristicChanged,
                 this, &QLowEnergyControllerPrivateAndroid::characteristicChanged);
+        connect(hub, &LowEnergyNotificationHub::serviceError,
+                this, &QLowEnergyControllerPrivateAndroid::serviceError);
     }
 }
 
@@ -500,7 +502,7 @@ void QLowEnergyControllerPrivateAndroid::servicesDiscovered(
 
     if (errorCode == QLowEnergyController::NoError) {
         //Android delivers all services in one go
-        const QStringList list = foundServices.split(QStringLiteral(" "), QString::SkipEmptyParts);
+        const QStringList list = foundServices.split(QStringLiteral(" "), Qt::SkipEmptyParts);
         for (const QString &entry : list) {
             const QBluetoothUuid service(entry);
             if (service.isNull())
@@ -549,7 +551,7 @@ void QLowEnergyControllerPrivateAndroid::serviceDetailsDiscoveryFinished(
         if (javaIncludes.isValid()) {
             const QStringList list = javaIncludes.toString()
                                                  .split(QStringLiteral(" "),
-                                                        QString::SkipEmptyParts);
+                                                        Qt::SkipEmptyParts);
             for (const QString &entry : list) {
                 const QBluetoothUuid service(entry);
                 if (service.isNull())
